const {app} = require('electron');
var process = require('process');
var path = require('path');
var fs = require('fs');
var childProcess = require('child_process');
const {BrowserWindow, dialog} = require('electron');

var appFolder = path.resolve(process.execPath, '..');
var rootFolder = path.resolve(appFolder, '..');
var updateDotExe = path.join(rootFolder, 'Update.exe');
var exeName = path.basename(process.execPath);

var Strings = require('./strings.js');
var strings = new Strings(app.getLocale());

var spawn = function(command, args, callback) {

    if (typeof callback !== 'function') {
        callback = function() {};
    }

    var stdout = '', spawnedProcess, error;

    try {
        spawnedProcess = childProcess.spawn(command, args);
    } catch (error) {
        process.nextTick(function() {
            callback(error, stdout);
        });
        return;
    }

    spawnedProcess.stdout.on('data', function(data) {
        stdout += data;
    });

    error = null;
    spawnedProcess.on('error', function(processError) {
        error = error || processError;
    });
    spawnedProcess.on('close', function(code, signal) {
        if (code !== 0) {
            error = error || new Error('Command failed: ' + (signal || code));
        }
        if (error) {
            error.code = code;
            error.stdout = stdout;
        }
        callback(error, stdout);
    });
};

var spawnUpdate = function(args, callback) {
    spawn(updateDotExe, args, callback);
};

var createShortcuts = function(callback) {
    spawnUpdate(['--createShortcut', exeName], callback);
};

var removeShortcuts = function(callback) {
    spawnUpdate(['--removeShortcut', exeName], callback);
};

var updateExistsSync = function() {
    return fs.existsSync(updateDotExe);
};

var restartApp = function() {
    app.on('will-quit', function() {
        spawnUpdate(['--processStart', exeName], function() {});
    });
    if (BrowserWindow.mainWindow) {
        BrowserWindow.mainWindow.forceClose = true;
    }
    app.quit();
};

var updateUrl = 'https://hedwig.bearychat.com/installer/' + process.arch;

var quitAndInstall = function() {
    if (updateExistsSync()) {
        restartApp();
    }
};

var downloadUpdate = function(callback) {

    spawnUpdate(['--download', updateUrl], function(error, stdout) {
        if (error) {
            callback(error);
            return;
        }

        var json, jsonObj, update;

        try {
            json = stdout.trim().split('\n').pop();
            jsonObj = JSON.parse(json);
            if (jsonObj && jsonObj.releasesToApply && jsonObj.releasesToApply.pop) {
                update = jsonObj.releasesToApply.pop();
            }
        } catch (error) {
            error.stdout = stdout;
            return callback(error);
        }

        return callback(null, update);
    });
};

var installUpdate = function(callback) {
    spawnUpdate(['--update', updateUrl], callback);
};

var silent = true;

var checkUpdate = function() {

    var onUpdateNotAvailable = function() {
        if (silent) {
            silent = false;
            return;
        }
        dialog.showMessageBox(BrowserWindow.mainWindow, {
            title: strings.translate('update-not-available'),
            message: strings.translate('update-not-available'),
            buttons: [strings.translate('ok')]
        });
    };

    if (!updateExistsSync()) {
        onUpdateNotAvailable();
        return;
    }

    downloadUpdate(function(error, update) {
        if (error) {
            onUpdateNotAvailable();
            return;
        }

        if (!update) {
            onUpdateNotAvailable();
            return;
        }

        installUpdate(function(error) {
            if (error) {
                onUpdateNotAvailable();
                return;
            }

            console.log('update-downloaded');

            var ret = dialog.showMessageBox(BrowserWindow.mainWindow, {
                title: strings.translate('update-downloaded'),
                message: update.version,
                detail: update.releaseNotes,
                buttons: [strings.translate('restart-now'), strings.translate('restart-later')]
            });

            if (ret === 0) {
                if (BrowserWindow.mainWindow) {
                    BrowserWindow.mainWindow.forceClose = true;
                }
                quitAndInstall();
            }
        });
    });
};

checkUpdate.handleEvents = function() {
    var squirrelCommand = process.argv[1];
    switch (squirrelCommand) {
        case '--squirrel-install':
            createShortcuts(function() {
                app.quit();
            });

            return true;
        case '--squirrel-updated':
            removeShortcuts(function() {
                createShortcuts(function() {
                    app.quit();
                });
            });

            return true;
        case '--squirrel-uninstall':
            // Undo anything you did in the --squirrel-install and
            // --squirrel-updated handlers

            removeShortcuts(function() {
                try {
                    childProcess.execSync('taskkill /IM BearyChat.exe /F');
                } catch (e) {
                    console.error(e);
                }

                app.quit();
            });

            return true;
        case '--squirrel-obsolete':
            // This is called on the outgoing version of your app before
            // we update to the new version - it's the opposite of
            // --squirrel-updated
            app.quit();
            return true;
        default:
            return false;
    }
};

module.exports = checkUpdate;
