'use strict';

const {app, dialog, session, BrowserWindow, ipcMain} = require('electron');
const path = require('path');

const Strings = require('./strings');
const strings = new Strings(app.getLocale());

let idCounter = 0;
const itemMap = new Map();
const restartUrlToPathMap = new Map();
const preference = require('./preference');

class Downloader {
    constructor() {
        session.defaultSession.on('will-download', this.handleWillDownload.bind(this));
        ipcMain.on('hedwig-download-cancel', this.handleCancel.bind(this));
        ipcMain.on('hedwig-download-restart', this.handleRestart.bind(this));
    }

    convertItem(item) {
        return {
            mime: item.getMimeType(),
            path: item.getSavePath(),
            url: item.getURL(),
            filename: item.getFilename(),
            totalBytes: item.getTotalBytes(),
            receivedBytes: item.getReceivedBytes()
        };
    }

    handleWillDownload(event, item, webContents) {
        const url = item.getURL();
        if (restartUrlToPathMap.has(url)) {
            item.setSavePath(restartUrlToPathMap.get(url));
            restartUrlToPathMap.delete(url);
        } else {
            const savePath = dialog.showSaveDialog(
                BrowserWindow.mainWindow,
                {
                    title: `${strings.translate('download-title')}${item.getFilename()}`,
                    defaultPath: path.resolve(preference.getDownloadLocation(), item.getFilename())
                }
            );

            if (!savePath) {
                event.preventDefault();
                return;
            }

            item.setSavePath(savePath);
        }

        const itemId = idCounter++;
        const itemToSend = Object.assign({
            id: itemId
        }, this.convertItem(item));
        webContents.send('hedwig-will-download', itemToSend);
        itemMap.set(itemId, item);

        item.on('updated', () => {
            if (!webContents || webContents.isDestroyed()) {
                item.cancel();
                return;
            }

            webContents.send('hedwig-download-item-update',
                Object.assign(itemToSend, this.convertItem(item)));
        });

        item.on('done', (e, state) => {
            if (!webContents || webContents.isDestroyed()) {
                return;
            }

            webContents.send('hedwig-download-item-done',
                Object.assign(itemToSend, this.convertItem(item)),
                state);
            itemMap.delete(itemId);
        });
    }

    handleCancel(event, teamId) {
        const item = itemMap.get(teamId);
        if (!item || item.isDestroyed()) {
            return;
        }
        item.cancel();
    }

    handleRestart(event, path, url) {
        const webContents = event.sender;
        webContents.downloadURL(url);
        restartUrlToPathMap.set(url, path);
    }
}

module.exports = function initDownloader() {
    return new Downloader();
};