(function (window, document, undefined) {
    'use strict';

    const {platform} = require('process');
    const {ipcRenderer} = require('electron');
    const {Menu, app} = require('electron').remote;

    /******* exported *******/
    var hedwig = window.hedwig || (window.hedwig = {});

    /******* constants *******/
    var TAB_ID_PREFIX, FRAME_ID_PREFIX;
    TAB_ID_PREFIX = 'tab-';
    FRAME_ID_PREFIX = 'frame-';

    var Strings = require("../strings.js");
    var strings = new Strings();
    strings.setLocale(app.getLocale());

    /******* DOM elements *******/
    var $teamTabs, $teamWebviews, $signinModal, $modalWebview, $loadingWrapper;
    $teamTabs = document.getElementById('team-tab');
    $teamWebviews = document.getElementById('team-webviews');
    $signinModal = document.getElementById('signin-modal');
    $modalWebview = document.getElementById('modal-webview');
    $loadingWrapper = document.getElementById('loading-wrapper');

    var pprTeamName = function (name) {
        var words = name.split(' ');
        var acronym;
        if (words.length > 1 && words[0][0].charCodeAt() < 256 &&
            words[1][0].charCodeAt() < 256) {
            acronym = words[0][0] + words[1][0];
        } else {
            acronym = words[0][0];
        }
        return acronym.toUpperCase();
    };
    var createTabItem = function (teamInfo) {
        var item = document.createElement('div');
        item.id = TAB_ID_PREFIX + teamInfo.id;
        item.className = 'team-tab-item';
        if (teamInfo.name) {
            item.innerHTML = pprTeamName(teamInfo.name);
        }
        item.addEventListener('click', function () {
            hedwig.nw.setCurrentTeam(teamInfo.id);
        });

        item.addEventListener('contextmenu', function() {
            var menu = Menu.buildFromTemplate([{
                label: strings.translate('remove-team'),
                click: function() {
                    hedwig.nw.signOut(teamInfo.id);
                }
            }]);
            menu.popup();
        })
        return item;
    };
    var createTeamWebview = function (teamInfo) {
        // FIXME: hack to enable node integration on <webview> create
        var div = document.createElement('div');
        div.innerHTML = '<webview preload="./preload.js"></webview>';
        var webview = div.querySelector('webview');

        webview.id = FRAME_ID_PREFIX + teamInfo.id;
        webview.src = teamInfo.url;

        hedwig.bindWebviewEvents(webview);

        return webview;
    };

    var switchTab = function () {
        toggleView($teamTabs, TAB_ID_PREFIX);
    };
    var switchWebview = function () {
        toggleView($teamWebviews, FRAME_ID_PREFIX);
    };
    var toggleView = function (node, prefix) {
        var children, i;
        children = node.children;
        for (i = 0; i < children.length; i++) {
            if (!hedwig.curTeamId || children[i].id === prefix + hedwig.curTeamId) {
                children[i].classList.add('selected');
            } else {
                children[i].classList.remove('selected');
            }
        }
    };

    function shortcutsHandler(e) {
        var modifier;
        if (platform === 'darwin') {
            modifier = e.metaKey;
        } else {
            modifier = e.ctrlKey;
        }
        if (!modifier) {
            return;
        }
        if (e.keyCode >= 49 && e.keyCode <= 56) {
            // 1-8
            hedwig.setNthTabToActive(e.keyCode - 49);
        }
        if (e.keyCode === 57) {
            hedwig.setLastTabToActive();
        }
        if (platform !== 'darwin' && e.keyCode === 82) {
            hedwig.reload();
        }
    }

    document.addEventListener('keydown', shortcutsHandler);


    hedwig.START_URL = 'https://bearychat.com/start';
    // hedwig.START_URL = 'http://stage.bearychat.com/start';
    // hedwig.START_URL = 'http://local.bearychat.com/start';
    hedwig.curTeamId = localStorage.current_team_id;
    hedwig.teams = JSON.parse(localStorage.teams || '{}');
    hedwig.teamUnreadMap = {};

    hedwig.updateTeamTab = function () {
        var tabId, tabItem, i, tabBarWidth;

        tabBarWidth = '64px';

        if (hedwig.teams.length < 2) {
            $teamTabs.style.display = 'none';
            $teamWebviews.style.marginLeft = '0';
            $loadingWrapper.style.left = '0';
        } else {
            $teamTabs.style.display = 'block';
            $teamWebviews.style.marginLeft = tabBarWidth;
            $loadingWrapper.style.left = tabBarWidth;
        }
        for (i in hedwig.teams) {
            tabId = TAB_ID_PREFIX + i;
            tabItem = document.getElementById(tabId);
            if (!tabItem) {
                tabItem = createTabItem(hedwig.teams[i]);
                $teamTabs.appendChild(tabItem);
            } else if (hedwig.teams[i].name) {
                tabItem.innerHTML = pprTeamName(hedwig.teams[i].name);
            }
        }
        switchTab();
    };

    hedwig.updateTeamWebviews = function () {
        var children, frameId, teamFrame, i, keys;
        children = $teamWebviews.children;
        keys = hedwig.teams.keys;
        if (keys.length === 1) {
            if (!children[0].id) {
                var team = hedwig.teams[keys[0]];
                children[0].id = FRAME_ID_PREFIX + team.id;
                children[0].src = team.url;
            }
        } else {
            for (i in hedwig.teams) {
                if (!children[0].id) {
                    children[0].remove();
                }
                frameId = FRAME_ID_PREFIX + i;
                teamFrame = document.getElementById(frameId);
                if (!teamFrame) {
                    teamFrame = createTeamWebview(hedwig.teams[i]);
                    $teamWebviews.appendChild(teamFrame);
                }
            }
        }
        switchWebview();
    };
    hedwig.setNthTabToActive = function (n) {
        var children = $teamTabs.children;
        if (children[n]) {
            hedwig.nw.setCurrentTeam(children[n].id.split('-')[1]);
        }
    };
    hedwig.setLastTabToActive = function () {
        var children = $teamTabs.children;
        hedwig.setNthTabToActive(children.length - 1);
    };

    hedwig.updateBadgeLabel = function () {
        var unread, i, tabId, tabEle;
        unread = false;
        for (i in hedwig.teamUnreadMap) {
            tabId = TAB_ID_PREFIX + i;
            tabEle = document.getElementById(tabId);
            if (!tabEle && !hedwig.teams[i]) {
                delete hedwig.teamUnreadMap[i];
                return;
            }
            if (hedwig.teamUnreadMap[i]) {
                unread = true;
                tabEle.classList.add('unread');
            } else {
                tabEle.classList.remove('unread');
            }
        }
        ipcRenderer.send('unread', unread);
    };

    var errorTimeoutId = null,
        cancelErrorTimeout = function () {
            if (errorTimeoutId != null) {
                clearTimeout(errorTimeoutId);
                errorTimeoutId = null;
            }
        },
        setErrorTimeout = function (timeout) {
            cancelErrorTimeout();
            errorTimeoutId = setTimeout(function () {
                hedwig.setLoadingError(true);
            }, timeout);
        },
        checkOnlineStatus = function () {
            if (!navigator.onLine) {
                // 假装 loading 一下然后显示出错
                setErrorTimeout(1000);
            } else {
                // 总也没 loading 进来然后显示出错
                setErrorTimeout(20 * 1000);
            }
        };

    hedwig.reload = function () {
        hedwig.setLoadingError(false);
        hedwig.setLoadingVisible(true);
        var frames = document.querySelectorAll('webview[src]');
        for (var i = 0; i < frames.length; i++) {
            frames[i].reload();
        }

        checkOnlineStatus();
    };

    var getFrameByTeamId = function (teamId) {
        var curFrameId;
        curFrameId = FRAME_ID_PREFIX + teamId;
        return document.querySelector(`webview[id="${curFrameId}"]`);
    };

    hedwig.focusTeam = function (teamId) {
        var curFrame = getFrameByTeamId(teamId);
        if (curFrame) {
            curFrame.focus();
        }
    };

    hedwig.blurTeam = function (teamId) {
        var curFrame = getFrameByTeamId(teamId);
        if (curFrame) {
            curFrame.blur();
        }
    };

    var isLoadingVisible = function () {
        return $loadingWrapper.style.display !== 'none';
    };

    var retryTimerId = null;
    var cancelRetry = function () {
        if (retryTimerId === null) {
            return;
        }
        clearTimeout(retryTimerId);
        retryTimerId = null;
    };
    var scheduleRetry = function () {
        cancelRetry();
        retryTimerId = setTimeout(function () {
            hedwig.reload();
        }, 5 * 1000);
    };

    hedwig.setLoadingVisible = function (visible) {
        if (visible) {
            $loadingWrapper.style.display = 'block';
        } else {
            $loadingWrapper.style.display = 'none';
            cancelRetry();
        }
    };

    hedwig.setLoadingError = function (error) {
        var className = 'error';
        if (error) {
            $loadingWrapper.classList.add(className);
            if (isLoadingVisible()) {
                scheduleRetry();
            }
        } else {
            $loadingWrapper.classList.remove(className);
            cancelRetry();
        }
    };

    hedwig.nw = (function () {
        return {
            setUnread: function (teamId) {
                hedwig.teamUnreadMap[teamId] = true;
                hedwig.updateBadgeLabel();
            },
            resetUnread: function (teamId) {
                hedwig.teamUnreadMap[teamId] = false;
                hedwig.updateBadgeLabel();
            },
            signIn: function (teamId, teamUrl) {
                console.log('sign in: ', teamId);
                if (!hedwig.teams[teamId]) {
                    hedwig.teams[teamId] = {
                        id: teamId,
                        url: teamUrl,
                        name: ''
                    };
                    hedwig.curTeamId = teamId;
                }
                if ($signinModal.style.display) {
                    hedwig.nw.closeSwitchModal();
                }
            },
            signOut: function (teamId) {
                console.log('sign out: ', teamId);
                delete hedwig.teams[teamId];
                delete hedwig.teamUnreadMap[teamId];
            },
            setTeamInfo: function (teamId, teamName) {
                var team = hedwig.teams[teamId];
                if (team) {
                    hedwig.teams[teamId] = Object.assign(team, {
                        name: teamName
                    });
                }
            },
            switchTeam: function () {
                if ($modalWebview.src !== hedwig.START_URL) {
                    $modalWebview.src = hedwig.START_URL;
                }
                $signinModal.style.display = 'block';
            },
            setCurrentTeam: function (teamId) {
                ipcRenderer.send('show');
                if (hedwig.curTeamId !== teamId) {
                    hedwig.blurTeam(hedwig.curTeamId);
                    hedwig.curTeamId = teamId;
                    switchTab();
                    switchWebview();
                    hedwig.focusTeam(teamId);
                }
            },
            screenshot: function () {
                ipcRenderer.send('screenshot');
            },
            showContextMenu: function () {
                console.log('context menu');
                ipcRenderer.send('context_menu');
            },
            closeSwitchModal: function () {
                $modalWebview.src = hedwig.START_URL;
                $signinModal.style.display = 'none';
                $modalWebview.src = '';
            },
            historyBack: function() {
                var webView = getFrameByTeamId(hedwig.curTeamId);
                if (webView) {
                    webView.goBack();
                }
            },
            historyForward: function() {
                var webView = getFrameByTeamId(hedwig.curTeamId);
                if (webView) {
                    webView.goForward();
                }
            }
        };
    })();

    const handleReady = () => {
        hedwig.setLoadingVisible(false);

        // fix https://github.com/bearyinnovative/hedwig/issues/135
        Object.keys(hedwig.teams).forEach(teamId => {
            const webview = getFrameByTeamId(teamId);
            if (webview) {
                if (teamId === hedwig.curTeamId) {
                    webview.blur();
                    webview.focus();
                } else {
                    webview.focus();
                    webview.blur();
                }
            }
        });
    };

    hedwig.handleEvent = (data) => {
        switch (data.type) {
            case 'set_unread':
                hedwig.nw.setUnread(data.team_id);
                break;
            case 'reset_unread':
                hedwig.nw.resetUnread(data.team_id);
                break;
            case 'sign_in':
                hedwig.nw.signIn(data.team_id, data.team_url);
                break;
            case 'sign_out':
                hedwig.nw.signOut(data.team_id);
                break;
            case 'set_team_info':
                hedwig.nw.setTeamInfo(data.team_id, data.team_name);
                break;
            case 'switch_team':
                hedwig.nw.switchTeam();
                break;
            case 'set_current_team':
                hedwig.nw.setCurrentTeam(data.team_id);
                break;
            case 'screenshot':
                hedwig.nw.screenshot();
                break;
            case 'ready':
                handleReady();
                break;
            case 'context_menu':
                hedwig.nw.showContextMenu();
                break;
        }
    };

    hedwig.bindWebviewEvents = (webview) => {
        webview.addEventListener('ipc-message', (event) => {
            if (event.channel !== 'hedwig') {
                return;
            }
            const [data] = event.args;
            hedwig.handleEvent(data);
        });

        webview.addEventListener('console-message', (event) => {
            const {level, message, line, sourceId} = event;
            console.log(level, message, line, sourceId);
        });

        webview.addEventListener('new-window', (event) => {
            ipcRenderer.send('new-window', event.url);
        });
    };

    window.addEventListener('message', function (event) {
        hedwig.handleEvent(event.data);
    });

    (function () {
        checkOnlineStatus();
        window.addEventListener('offline', function () {
            hedwig.setLoadingError(true);
        });

        var $reconnectButton = document.getElementById('reconnect-button');
        $reconnectButton.addEventListener('click', function () {
            hedwig.reload();
        }, false);
    })();

    hedwig.teams = new Proxy(hedwig.teams, {
        deleteProperty(target, prop) {
            const ret = delete target[prop];
            if (!ret) {
                return ret;
            }
            const teamId = prop;
            const keys = Object.keys(target);
            console.log('remaining teams', keys);
            hedwig.curTeamId = keys[0];
            if (keys.length > 0) {
                document.getElementById(TAB_ID_PREFIX + teamId).remove();
                document.getElementById(FRAME_ID_PREFIX + teamId).remove();
            } else {
                const $webview = document.querySelector('webview');
                $webview.id = '';
                const handleStartLoading = () => {
                    $webview.removeEventListener('did-start-loading', handleStartLoading);
                    $webview.src = hedwig.START_URL;
                };
                $webview.addEventListener('did-start-loading', handleStartLoading);
            }

            hedwig.updateTeamTab();
            hedwig.updateTeamWebviews();
            return ret;
        },

        set(target, prop, value) {
            target[prop] = value;
            hedwig.updateTeamTab();
            hedwig.updateTeamWebviews();
            return true;
        },

        get(target, prop) {
            if (prop === 'length') {
                return Object.keys(target).length;
            }

            if (prop === 'keys') {
                return Object.keys(target);
            }

            return target[prop];
        },

        has(target, prop) {
            return prop in target;
        }
    });

    module.exports = hedwig;
})(window, document);
