(function () {
    let mainWindow = null;
    const {app} = require('electron');

    const shouldQuit = app.makeSingleInstance(() => {
        if (mainWindow) {
            mainWindow.show();
            if (mainWindow.isMinimized()) {
                mainWindow.restore();
            }
            mainWindow.focus();
        }
    });

    if (shouldQuit) {
        app.quit();
        return;
    }

    var process = require('process');
    var platform = process.platform;

    var checkUpdate = function () {
    };

    if (platform === 'win32') {
        checkUpdate = require('./check_update_win');

        if (checkUpdate.handleEvents()) {
            return;
        }
    } else if (platform === 'darwin') {
        checkUpdate = require('./check_update_osx');
    }

    var manifest = require('./package.json');

    const {
        BrowserWindow, globalShortcut, Tray, Menu,
        shell, ipcMain, nativeImage
    } = require('electron');
    var path = require('path');
    var trayIcons = require('./tray.json')[platform];
    var fs = require('fs');
    var nodeUrl = require('url');
    var mkdirp = require('mkdirp');

    var createWindow = require('./create-window');
    var Utils = require('./utils');

    var windowState = null;
    var tray = null;

    var dataDir = path.resolve(app.getPath('appData'), manifest.name, 'Custom');
    var statePath = path.resolve(dataDir, 'winstate');

    mkdirp(dataDir);

    var showMainWindow = function () {
        if (!mainWindow) {
            return;
        }
        mainWindow.show();
        mainWindow.focus();
    };

    var imageCache = {};

    var Strings = require("./strings.js");
    var strings = new Strings();

    ['normal', 'alt_normal', 'unread', 'alt_unread'].forEach(function (key) {
        'use strict';
        imageCache[key] = nativeImage.createFromPath(
            path.resolve(__dirname, trayIcons[key]));
    });

    var isValidNumber = function (n) {
        return (typeof n === 'number') && !isNaN(n);
    };

    var initWindowState = function () {
        'use strict';
        try {
            windowState = JSON.parse(fs.readFileSync(statePath, {
                encoding: 'utf8',
            }));
        } catch (e) {
            windowState = {};
        }

        if (!windowState.bounds) {
            windowState.bounds = {};
        }

        var bounds = windowState.bounds;

        const {screen: electronScreen} = require('electron');
        var size = electronScreen.getPrimaryDisplay().workAreaSize;

        var DEFAULT_WIDTH = 800;
        var DEFAULT_HEIGHT = 600;

        if (!isValidNumber(bounds.x) || !isValidNumber(bounds.y) || !isValidNumber(bounds.width) || !isValidNumber(bounds.height)) {

            // Bounds not valid, set to default
            bounds.x = Math.round((size.width - DEFAULT_WIDTH) / 2);
            bounds.y = Math.round((size.height - DEFAULT_HEIGHT) / 2);
            bounds.width = DEFAULT_WIDTH;
            bounds.height = DEFAULT_HEIGHT;
        }

        var currentDisplay = electronScreen.getDisplayMatching(bounds),
            currentRect = currentDisplay.workArea;

        // Check if bounds in screen
        if (bounds.x < currentRect.x || bounds.x > currentRect.x + currentRect.width) {
            bounds.x = currentRect.x;
        }

        if (bounds.y < currentRect.y || bounds.y > currentRect.y + currentRect.height) {
            bounds.y = currentRect.y;
        }

        if (bounds.width > currentRect.width) {
            bounds.width = currentRect.width;
        }

        if (bounds.height > currentRect.height) {
            bounds.height = currentRect.height;
        }
    };

    var savingState = false;

    var saveWindowState = function () {
        'use strict';

        if (savingState) {
            return;
        }

        savingState = true;

        fs.writeFile(statePath, JSON.stringify(windowState), function (err) {
            if (err) {
                console.error(err);
                return;
            }
            savingState = false;
        });

        console.log(windowState);
    };

    const newWindowOpenLinkCallback = (event, url) => {
        event.preventDefault();
        shell.openExternal(url);
    };

    const handleInternalUrls = (url) => {
        const urlObj = nodeUrl.parse(url);
        if (urlObj.host.endsWith('bearychat.com') && urlObj.path.startsWith('/snippet/=')) {
            const win = createWindow(url, {
                width: 900,
                height: 750
            });
            win.webContents.on('new-window', newWindowOpenLinkCallback);
            win.webContents.on('will-navigate', newWindowOpenLinkCallback);
            return true;
        }
        return false;
    };

    const handleNewWindow = (event, url) => {
        event.preventDefault();
        if (handleInternalUrls(url)) {
            return;
        }
        shell.openExternal(url);
    };

    var initApp = function () {
        'use strict';

        initWindowState();

        BrowserWindow.mainWindow = mainWindow = new BrowserWindow({
            x: windowState.bounds.x,
            y: windowState.bounds.y,
            width: windowState.bounds.width,
            height: windowState.bounds.height
        });
        mainWindow.loadURL('file://' + path.resolve(__dirname, 'fe/index.html'));
        mainWindow.forceClose = false;

        mainWindow.on('close', function (event) {
            if (!mainWindow.forceClose) {
                if (mainWindow.isFullScreen()) {
                    mainWindow.once('leave-full-screen', function () {
                        mainWindow.hide();
                    });
                    console.log(mainWindow.once);
                    mainWindow.setFullScreen(false);
                } else {
                    mainWindow.hide();
                }
                event.preventDefault();
            }
        });

        app.on('before-quit', function () {
            mainWindow.forceClose = true;
            saveWindowState();
        });

        app.on('activate', function (hasVisibleWindows) {
            showMainWindow();
        });

        mainWindow.webContents.on('new-window', handleNewWindow);

        mainWindow.webContents.on('did-finish-load', function () {
            mainWindow.webContents.loadFinished = true;
        });

        mainWindow.on('resize', function () {
            windowState.bounds = mainWindow.getBounds();
            saveWindowState();
        });

        mainWindow.on('move', function () {
            windowState.bounds = mainWindow.getBounds();
            saveWindowState();
        });
    };

    var sendToWebContents = function () {
        'use strict';

        var wc = mainWindow.webContents;
        if (wc.loadFinished) {
            wc.send.apply(wc, arguments);
        }
    };


    var initTray = function () {
        'use strict';

        tray = new Tray(imageCache.normal);
        tray.setPressedImage(imageCache.alt_normal);
        tray.setToolTip(manifest.name);

        if (platform !== 'darwin') {
            var menu = Menu.buildFromTemplate([{
                label: strings.translate('open-bearychat'),
                click: showMainWindow
            }, {
                type: 'separator'
            }, {
                label: strings.translate('switch-team'),
                accelerator: 'CommandOrControl + N',
                click: switchTeam
            }, {
                type: 'separator'
            }, {
                label: strings.translate('check-update'),
                click: function () {
                    checkUpdate();
                },
                visible: platform === 'win32'
            }, {
                label: strings.translate('about-bearychat'),
                click: showAbout
            }, {
                type: 'separator'
            }, {
                label: strings.translate('quit-bearychat'),
                click: function () {
                    app.quit();
                }
            }]);
            tray.setContextMenu(menu);
        }
        tray.on('click', showMainWindow);
    };

    var switchTeam = function () {
        'use strict';
        showMainWindow();
        sendToWebContents('switch-team');
    };

    var showAbout = function () {
        'use strict';
        require('./window/about').show();
    };

    var globalActivateShortcut = 'Ctrl + Super + X';

    var initMenu = function () {
        'use strict';
        if (manifest.debug) {
            template[template.length - 1].submenu.push({
                label: 'Toggle DevTools',
                accelerator: 'Alt+CommandOrControl+I',
                click: function () {
                    var currentWindow = BrowserWindow.getFocusedWindow() || mainWindow;
                    currentWindow.toggleDevTools();
                }
            });
        }
        if (platform === 'darwin') {
            var menu = Menu.buildFromTemplate(require('./menu/menu-darwin'));
        } else if (platform === 'win32') {
            var menu = Menu.buildFromTemplate(require('./menu/menu-windows'));
        } else {
            var menu = Menu.buildFromTemplate(require('./menu/menu-linux'));
        }
        Menu.setApplicationMenu(menu);
    };

    var initEvents = function () {
        'use strict';

        var setDockBadge;

        if (platform === 'darwin') {
            setDockBadge = function (unread) {
                if (unread) {
                    app.dock.setBadge('•');
                } else {
                    app.dock.setBadge('');
                }
            };
        } else {
            setDockBadge = function () {
            };
        }

        ipcMain.on('unread', function (event, unread) {
            setDockBadge(unread);

            if (unread) {
                tray.setImage(imageCache.unread);
                tray.setPressedImage(imageCache.alt_unread);
            } else {
                tray.setImage(imageCache.normal);
                tray.setPressedImage(imageCache.alt_normal);
            }
        });

        ipcMain.on('show', showMainWindow);

        var popupMenu = Menu.buildFromTemplate([{
            label: strings.translate('copy'),
            role: 'copy'
        }, {
            label: strings.translate('paste'),
            role: 'paste'
        }]);

        ipcMain.on('context_menu', function () {
            popupMenu.popup(mainWindow);
        });

        ipcMain.on('new-window', handleNewWindow);

        mainWindow.on('focus', function () {
            sendToWebContents('focus');
        });

        mainWindow.on('blur', function () {
            sendToWebContents('blur');
        });

        if (platform === 'win32') {
            var logo = new nativeImage.createFromPath(path.resolve(__dirname, 'fe/images/logo.png'));

            let lastNotificationWebContents = null;

            tray.on('balloon-click', function () {
                lastNotificationWebContents && lastNotificationWebContents.send('notification-clicked');
            });

            ipcMain.on('notification', function (event, data) {
                lastNotificationWebContents = event.sender;
                tray.displayBalloon({
                    icon: logo,
                    title: data.title,
                    content: data.content
                });
            });
        }

        var globalActivateReg = globalShortcut.register(globalActivateShortcut, showMainWindow);
        if (!globalActivateReg) {
            console.warn('Activation global shortcut registration failed.');
        }
    };

    var registerScreenshot = () => {
        if (platform === 'darwin' || platform === 'win32') {
            const Screenshot = require('./screenshot');
            const preference = require('./preference');
            var screenshotKey = preference.getScreenshotKey();
            if (screenshotKey) {
                var screenshotReg = globalShortcut.register(screenshotKey, Screenshot.screenshot);
                if (!screenshotReg) {
                    console.warn('Screenshot global shortcut registration failed.');
                }
            }
        }
    };

    var main = () => {
        'use strict';

        strings.setLocale(app.getLocale());
        checkUpdate();

        initApp();
        initTray();
        initMenu();
        initEvents();
        registerScreenshot();

        require('./downloader')();
        require('./ipc-main').init();
    };

    app.on('ready', main);

    module.exports = main;

})();
